# regressors
from .gp_regressor import GPRegressor
from .feedforward_regressor import FeedforwardRegressor
from .transformer_regressor import TransformerRegressor
from .transformer_fft_regressor import TransformerFFTRegressor
from .mamba_regressor import MambaRegressor
from .mamba_fft_regressor import MambaFFTRegressor
from .gp_head_regressor import GPHeadRegressor

# classifiers
from .feedforward_image_classifier import FeedforwardImageClassifier
from .transformer_image_classifier import TransformerImageClassifier
from .transformer_fft_image_classifier import TransformerFFTImageClassifier
from .mamba_image_classifier import MambaImageClassifier
from .mamba_fft_image_classifier import MambaFFTImageClassifier
from .cnn_image_classifier import CNNImageClassifier



from catasta import Foundation, Dataset, Scaffold
from catasta.models import GPRegressor


def objective(params: dict) -> float:
    model = GPRegressor(
        n_inducing_points=params["n_inducing_points"],
        n_inputs=1,
        kernel=params["kernel"],
        mean=params["mean"],
    )

    dataset_root: str = "data/incomplete/"
    dataset = Dataset(
        dataset_root,
        task="regression",
        input_name="input",
        output_name="output",
    )

    scaffold = Scaffold(
        model=model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="variational_elbo",
        verbose=False,
    )

    scaffold.train(
        epochs=100,
        batch_size=32,
        lr=1e-3,
    )

    info = scaffold.evaluate()

    return info.rmse


def main() -> None:
    hp_space = {
        "n_inducing_points": (16, 32),
        "kernel": ("matern", "rbf"),
        "mean": ("constant", "zero"),
    }

    foundation = Foundation(
        hyperparameter_space=hp_space,
        objective_function=objective,
        sampler="bogp",
        n_trials=10,
        direction="minimize",
        catch_exceptions=True,
    )

    optimization_info = foundation.optimize()

    print(optimization_info)


if __name__ == "__main__":
    main()



from typing import NamedTuple

from datetime import datetime


class TrialInfo(NamedTuple):
    """Dataclass containing the trial information.

    Attributes
    ----------
    trial_number : int
        Number of the trial.
    hyperparameters : dict
        Hyperparameters of the trial.
    metric : float
        Metric value of the trial.
    datetime_start : datetime
        Start datetime of the trial.
    datetime_end : datetime
        End datetime of the trial.
    duration : float
        Duration of the trial.
    """
    number: int
    hyperparameters: dict
    metric: float
    datetime_start: datetime
    datetime_end: datetime
    duration: float



from optuna.samplers import (
    BaseSampler,
    RandomSampler,
    TPESampler,
    GPSampler,
)


def get_sampler(sampler: str) -> BaseSampler:
    match sampler.lower():
        case "random" | "rs":
            return RandomSampler()
        case "tpe":
            return TPESampler()
        case "gp" | "bogp":
            return GPSampler()
        case _:
            raise ValueError("Invalid sampler")



import platform
import time
import os
import gc

import numpy as np

import torch
from torch import Tensor
from torch.nn import Module
from torch.optim import Optimizer
from torch.utils.data import DataLoader
from torch.distributions import Distribution

from gpytorch.mlls import MarginalLogLikelihood

from ..dataset import CatastaDataset
from .utils import (
    get_optimizer,
    get_loss_function,
    ModelStateManager,
    TrainingLogger,
)
from ..dataclasses import TrainInfo, EvalInfo
from ..log import ansi


def _get_device(device: str) -> torch.device:
    match device:
        case "cpu":
            return torch.device("cpu")
        case "cuda":
            return torch.device("cuda")
        case "auto":
            return torch.device("cuda" if torch.cuda.is_available() else "cpu")
        case _:
            raise ValueError("Invalid device")


class Scaffold:
    def __init__(self, *,
                 model: Module,
                 dataset: CatastaDataset,
                 optimizer: str | Optimizer,
                 loss_function: str | Module,
                 device: str = "auto",
                 verbose: bool = True,
                 ) -> None:
        """The `Scaffold` class is a high-level API for training models on Catasta datasets.

        Arguments
        ---------
        model : torch.nn.Module
            The model to be trained.
        dataset : ~catasta.datasets.CatastaDataset
            The dataset to be used for training, validation, and testing.
        optimizer : str | torch.optim.Optimizer
            The optimizer to be used for training. The user can either pass a string with the name of the optimizer or a custom optimizer. One of: adam, sgd, adamw, lbfgs, rmsprop, rprop, adadelta, adagrad, adamax, asgd, sparseadam.
        loss_function : str | torch.nn.Module
            The loss function to be used for training. The user can either pass a string with the name of the loss function or a custom loss function. A list of available loss functions is in ~catasta.scaffolds.utils.available_loss_functions
        device : str, optional
            The device to be used for training. One of "cpu", "cuda", or "auto". Defaults to "auto".
        verbose : bool, optional
            Whether to print training information or not. Defaults to True.
        """
        self.task: str = dataset.task

        self.device: torch.device = _get_device(device)
        self.input_dtype: torch.dtype = torch.float32
        self.output_dtype: torch.dtype = torch.float32 if self.task == "regression" else torch.long

        self.verbose: bool = verbose

        self.model: Module = model.to(self.device, self.input_dtype)

        self.dataset: CatastaDataset = dataset

        self.optimizer_id: str | Optimizer = optimizer
        self.loss_function_id: str | Module = loss_function

        if not self.verbose:
            return

        n_params: int = sum(p.numel() for p in self.model.parameters() if p.requires_grad)
        device_name: str = torch.cuda.get_device_name() if self.device.type == "cuda" else platform.processor()
        n_samples: int = len(self.dataset.train)  # type: ignore
        print(
            f"{ansi.BOLD}{ansi.GREEN}-> training info\n{ansi.RESET}"
            f"   |> task:    {self.task}\n"
            f"   |> dataset: {self.dataset.root} ({n_samples} samples)\n"
            f"   |> model:   {self.model.__class__.__name__} ({n_params} parameters)\n"
            f"   |> device:  {self.device} ({device_name})"
        )

    def train(self, *,
              epochs: int,
              batch_size: int,
              lr: float,
              early_stopping_alpha: float | None = None,
              shuffle: bool = True,
              data_loader_workers: int = 0,
              ) -> TrainInfo:
        """Train the model on the dataset.

        Arguments
        ---------
        epochs: int
            The number of epochs to train the model.
        batch_size: int
            The batch size to use for training.
        lr: float
            The initial learning rate.
        early_stopping_alpha: bool, optional
            The smoothing factor for the early stopping criterion. If set to a value, the training will stop when the validation loss curve starts increasing. A good smoothing factor ranges in [0.95, 1). Defaults to None.
        shuffle: bool, optional
            Whether to shuffle the training data or not . Defaults to True.
        data_loader_workers: int, optional
            The number of workers to use for the data loader. Defaults to 0. Tip: set to 4 times the number of GPUs.
        """
        # VARIABLES
        optimizer: Optimizer = get_optimizer(self.optimizer_id, self.model, lr)
        loss_function: Module = get_loss_function(self.loss_function_id, self.model, len(self.dataset.train))  # type: ignore

        model_state_manager: ModelStateManager = ModelStateManager(early_stopping_alpha)

        training_logger: TrainingLogger = TrainingLogger(self.task, epochs)

        train_data_loader: DataLoader = DataLoader(
            self.dataset.train,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=data_loader_workers,
            pin_memory=True if self.device.type == "cuda" else False,
        )
        val_data_loader: DataLoader = DataLoader(
            self.dataset.validation,
            batch_size=batch_size,
            shuffle=False,
            num_workers=data_loader_workers,
            pin_memory=True if self.device.type == "cuda" else False,
        )

        gc.collect()
        torch.cuda.empty_cache()

        # TRAINING LOOP
        for epoch in range(epochs):
            try:
                start_time: float = time.time()

                # train
                self.model.train()
                train_loss = self._epoch(train_data_loader, loss_function, optimizer)

                # validation
                self.model.eval()
                with torch.no_grad():
                    val_loss = self._epoch(val_data_loader, loss_function, None)

                model_state_manager(self.model, val_loss)

                if model_state_manager.stop:
                    print(
                        f"\n{ansi.BOLD}{ansi.YELLOW}-> early stopping{ansi.RESET}\n"
                        f"   |> epoch: {epoch + 1}"
                    )
                    break

                training_logger.log(
                    train_loss=train_loss,
                    val_loss=val_loss,
                    lr=optimizer.param_groups[0]["lr"],
                    epoch=epoch + 1,
                    time_per_epoch=time.time() - start_time,
                )

                print(training_logger) if self.verbose else None

            except KeyboardInterrupt:
                print(
                    f"\n{ansi.BOLD}{ansi.YELLOW}-> training interrupted{ansi.RESET}\n"
                    f"   |> epoch: {epoch + 1}"
                )
                break

        # END OF TRAINING
        train_info = training_logger.get_info()

        model_state_manager.load_best_model_state(self.model)

        return train_info

    def _epoch(self,
               data_loader: DataLoader,
               loss_function: Module,
               optimizer: Optimizer | None,
               ) -> float:
        cumulated_loss: Tensor = torch.tensor(0.0, device=self.device)
        total_samples: int = 0

        for inputs, targets in data_loader:
            inputs = inputs.to(self.device, self.input_dtype, non_blocking=True)
            targets = targets.to(self.device, self.output_dtype, non_blocking=True)

            if optimizer is not None:
                optimizer.zero_grad(set_to_none=True)

            output = self.model(inputs)

            loss: Tensor = loss_function(output, targets)  # type: ignore
            if isinstance(loss_function, MarginalLogLikelihood):
                loss = -loss

            if optimizer is not None:
                loss.backward()
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), 1.0)  # type: ignore
                optimizer.step()

            cumulated_loss += loss.detach() * inputs.shape[0]
            total_samples += inputs.shape[0]

        epoch_loss = cumulated_loss / total_samples if total_samples > 0 else torch.tensor(0.0, device=self.device)

        return epoch_loss.item()

    @torch.no_grad()
    def evaluate(self, batch_size: int | None = None) -> EvalInfo:
        """Evaluate the model on the test set of the dataset.

        Arguments
        ---------
        batch_size: int, optional
            The batch size to use for evaluation. If None, it uses the entire test set. Defaults to None.

        Returns
        -------
        ~catasta.dataclasses.EvalInfo
            The evaluation information.
        """
        self.model.eval()

        batch_size = len(self.dataset.test) if batch_size is None else batch_size  # type: ignore
        data_loader: DataLoader = DataLoader(
            self.dataset.test,
            batch_size=batch_size,
            shuffle=False,
            num_workers=0,
            pin_memory=True if self.device.type == "cuda" else False,
        )

        true_output = []
        predicted_output = []
        predicted_output_std = []
        for x, y in data_loader:
            x = x.to(self.device, self.input_dtype)

            output = self.model(x)
            true_output.append(y.cpu().numpy())

            if isinstance(output, Distribution):
                predicted_output.append(output.mean.cpu().numpy())
                predicted_output_std.append(output.stddev.cpu().numpy())
            else:
                predicted_output.append(output.cpu().numpy())
                predicted_output_std.append(np.zeros_like(output.cpu().numpy()))

        return EvalInfo(
            task=self.task,
            true_output=np.concatenate(true_output),
            predicted_output=np.concatenate(predicted_output),
            predicted_std=np.concatenate(predicted_output_std),
        )

    def save(self, path: str) -> None:
        """Save the model to a file.

        Arguments
        ---------
        path : str
            The path where to save the model. If the path is a directory, the model is saved as the model class name in the directory. If the path ends with `.onnx`, the model is saved in ONNX format.

        Raises
        ------
        ValueError
            If the path is not a directory, a `.pt` or `.onnx` file.
        """

        if "." in path and not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
            print(
                f"{ansi.BOLD}{ansi.YELLOW}-> specified path does not exist{ansi.RESET}\n"
                f"   |> created directory: {os.path.dirname(path)}"
            )

        if "." not in path and not os.path.exists(path):
            os.makedirs(path)
            print(
                f"{ansi.BOLD}{ansi.YELLOW}-> specified path does not exist{ansi.RESET}\n"
                f"   |> created directory: {path}"
            )

        self.model.to("cpu")

        if os.path.isdir(path):
            path = os.path.join(path, f"{self.model.__class__.__name__}.pt")

        if path.endswith(".pt") or path.endswith(".pth"):
            torch.save(self.model, path)

        elif path.endswith(".onnx"):
            # check if the onnx library is installed
            try:
                import onnx  # type: ignore
            except ImportError:
                print(
                    f"{ansi.BOLD}{ansi.RED}-> onnx library not found{ansi.RESET}\n"
                    f"   |> onnx is a special use case, please install it manually\n"
                    f"   |> the model will be saved as a `.pt` file\n"
                    f"   |> run: pip install onnx"
                )

                torch.save(self.model, path.replace(".onnx", ".pt"))

                print(
                    f"{ansi.BOLD}{ansi.GREEN}-> model saved{ansi.RESET}\n"
                    f"   |> path: {path.replace('.onnx', '.pt')}"
                )
                return

            input_shape = self.dataset.train[0][0].shape
            dummy_input = torch.randn(1, *input_shape, dtype=self.input_dtype)

            torch.onnx.export(
                self.model,
                dummy_input,  # type: ignore
                path,
                export_params=True,
                opset_version=11,
                do_constant_folding=True,
                input_names=["input"],
                output_names=["output"],
                dynamic_axes={
                    "input": {0: "batch_size"},
                    "output": {0: "batch_size"}
                }
            )
        else:
            raise ValueError("Invalid path. Must be a directory, a `.pt` or `.onnx` file.")

        print(
            f"{ansi.BOLD}{ansi.GREEN}-> model saved{ansi.RESET}\n"
            f"   |> path: {path}"
        )



import torch
from torch import Tensor, Size
from torch.nn import Module

from gpytorch.models import ApproximateGP
from gpytorch.variational import (
    CholeskyVariationalDistribution,
    VariationalStrategy,
    IndependentMultitaskVariationalStrategy,
)
from gpytorch.means import (
    ZeroMean,
    ConstantMean,
    Mean,
)
from gpytorch.kernels import (
    Kernel,
    ScaleKernel,
    RBFKernel,
    MaternKernel,
    RQKernel,
    RFFKernel,
    PeriodicKernel,
    LinearKernel,
)
from gpytorch.distributions import MultivariateNormal
from gpytorch.likelihoods import (
    Likelihood,
    GaussianLikelihood,
    BernoulliLikelihood,
    LaplaceLikelihood,
    SoftmaxLikelihood,
    StudentTLikelihood,
    BetaLikelihood,
    MultitaskGaussianLikelihood,
)


def _get_kernel(id: str, n_inputs: int, use_ard: bool, batch_shape: Size) -> Kernel:
    ard_num_dims = n_inputs if use_ard else None
    match id.lower():
        case "rq":
            return RQKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case "matern":
            return MaternKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case "rbf":
            return RBFKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case "rff":
            return RFFKernel(num_samples=n_inputs, batch_shape=batch_shape)
        case "periodic":
            return PeriodicKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case "linear":
            return LinearKernel(ard_num_dims=ard_num_dims, batch_shape=batch_shape)
        case _:
            raise ValueError(f"Unknown kernel: {id}")


def _get_mean_module(id: str, batch_shape: Size) -> Mean:
    match id.lower():
        case "constant":
            return ConstantMean(batch_shape=batch_shape)
        case "zero":
            return ZeroMean(batch_shape=batch_shape)
        case _:
            raise ValueError(f"Unknown mean: {id}")


def _get_likelihood(id: str, n_outputs: int) -> Likelihood:
    if n_outputs > 1:
        return MultitaskGaussianLikelihood(num_tasks=n_outputs)

    match id.lower():
        case "gaussian":
            return GaussianLikelihood()
        case "bernoulli":
            return BernoulliLikelihood()
        case "laplace":
            return LaplaceLikelihood()
        case "softmax":
            return SoftmaxLikelihood()
        case "studentt":
            return StudentTLikelihood()
        case "beta":
            return BetaLikelihood()
        case _:
            raise ValueError(f"Unknown likelihood: {id}")


class GPHeadRegressor(ApproximateGP):
    def __init__(self, *,
                 pre_model: Module,
                 pre_model_output_dim: int,
                 n_inputs: int,
                 n_outputs: int,
                 n_inducing_points: int,
                 kernel: str = "rq",
                 mean: str = "constant",
                 likelihood: str = "gaussian",
                 use_ard: bool = True,
                 ) -> None:
        '''
        Arguments
        ---------
        pre_model: Module
            The model that will be used to preprocess the input data
        pre_model_output_dim: int
            The output dimension of the pre_model
        n_inducing_points: int
            Number of inducing points
        n_inputs: int
            Number of input dimensions
        n_outputs: int
            Number of output dimensions
        kernel: str
            Kernel to use. One of "rq", "matern", "rbf", "rff", "periodic", "linear"
        mean: str
            Mean to use. One of "constant", "zero"
        likelihood: str
            Likelihood to use. One of "gaussian", "bernoulli", "laplace", "softmax", "studentt", "beta"
        use_ard: bool
            Whether to use Automatic Relevance Determination
        '''
        if n_outputs > 1:
            raise NotImplementedError("Multi-output regression is not yet supported")

        batch_shape = Size([]) if n_outputs == 1 else Size([n_outputs])
        inducing_points_shape = (n_inducing_points, n_inputs) if n_outputs == 1 else (n_outputs, n_inducing_points, n_inputs)
        inducing_points: Tensor = torch.randn(inducing_points_shape, dtype=torch.float32)

        variational_distribution = CholeskyVariationalDistribution(
            n_inducing_points,
            batch_shape=batch_shape,
        )
        base_variational_strategy = VariationalStrategy(
            self,
            inducing_points,
            variational_distribution,
            learn_inducing_locations=True,
        )
        variational_strategy = base_variational_strategy if n_outputs == 1 else IndependentMultitaskVariationalStrategy(
            base_variational_strategy=base_variational_strategy,
            num_tasks=n_outputs,
        )

        super().__init__(variational_strategy)
        self.mean_module = _get_mean_module(mean, batch_shape)
        self.covar_module = ScaleKernel(
            _get_kernel(
                kernel,
                pre_model_output_dim,
                use_ard,
                batch_shape
            ),
            batch_shape=batch_shape,
        )
        self.likelihood = _get_likelihood(likelihood, n_outputs)
        self.pre_model = pre_model

    def forward(self, x: Tensor) -> MultivariateNormal:
        x = self.pre_model(x)

        mean_x: Tensor = self.mean_module(x)  # type: ignore
        covar_x: Tensor = self.covar_module(x)  # type: ignore

        return MultivariateNormal(mean_x, covar_x)



from typing import NamedTuple
import numpy as np

from ..log import ansi


class TrainInfo(NamedTuple):
    """NamedTuple for storing training information.

    Attributes
    ----------
    train_losses : np.ndarray
        Array of the training losses for each epoch.
    best_train_loss : float
        The minimum value of the training losses.
    val_losses : np.ndarray
        Array of the validation losses for each epoch.
    best_val_loss : float
        The minimum value of the validation losses.
    lr_values : np.ndarray
        Array of the learning rates for each epoch.
    """
    train_losses: np.ndarray
    best_train_loss: float

    val_losses: np.ndarray
    best_val_loss: float

    lr_values: np.ndarray

    def __repr__(self) -> str:
        return f"\n{ansi.BOLD}{ansi.BLUE}-> training results{ansi.RESET}\n" \
            f"   |> best train loss: {self.best_train_loss:.4f}\n" \
            f"   |> best val loss:   {self.best_val_loss:.4f}"



from torch.nn import Module


class ModelStateManager:
    def __init__(self, alpha: float | None) -> None:
        self.early_stopping = True if alpha is not None else False

        self.best_loss: float = float("inf")
        self.best_model_state: dict = {}

        self.alpha: float = alpha if alpha is not None else 0.0
        self.stop: bool = False

        self.prev_loss: float = float("inf")
        self.derivative: float = 0.0

    def __call__(self, model: Module, loss: float) -> None:
        if not self.best_model_state:
            self.best_model_state = model.state_dict()

        if loss < self.best_loss:
            self.best_loss = loss
            self.best_model_state = model.state_dict()

        if not self.early_stopping:
            return

        unfiltered_derivate: float = loss - self.prev_loss if self.prev_loss != float("inf") else 0.0
        self.derivative = self.alpha * self.derivative + (1 - self.alpha) * unfiltered_derivate

        self.prev_loss = loss

        if self.derivative > 0:
            self.stop = True

    def load_best_model_state(self, model: Module) -> None:
        model.load_state_dict(self.best_model_state)



from catasta import Scaffold, Dataset
from catasta.models import TransformerFFTImageClassifier
from catasta.dataclasses import EvalInfo


def main() -> None:
    model = TransformerFFTImageClassifier(
        input_shape=(28, 28, 3),
        n_classes=10,
        n_patches=7,
        d_model=16,
        n_heads=2,
        n_layers=2,
        feedforward_dim=16,
        head_dim=16,
        dropout=0.1,
    )

    dataset = Dataset(
        "data/reduced_mnist",
        task="classification",
    )

    scaffold = Scaffold(
        model=model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="cross_entropy",
    )

    scaffold.train(
        epochs=100,
        batch_size=128,
        lr=1e-3,
    )

    eval_info: EvalInfo = scaffold.evaluate(batch_size=32)
    print(eval_info)


if __name__ == '__main__':
    main()
